/*
 * Copyright (C) 2012-2013, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"

/*
 * statvfs() on GNU/Linux may call stat() internally, and stat() is bad
 * for network mounts which may be stalled/slow, so favor the non-portable
 * statfs() on GNU/Linux
 */
#ifdef __linux__
#include <sys/vfs.h>
#define MY_STATFS statfs
#else /* statvfs() is POSIX */
#define MY_STATFS statvfs
#endif

/*
 * prevents us from using filesystems of unknown size, since those could
 * be stalled/dead network mounts
 */
bool mog_fs_usable(const char *path)
{
	struct MY_STATFS buf;

retry:
	errno = 0;
	if (MY_STATFS(path, &buf) == 0)
		return (buf.f_blocks > 0);

	/* unknown */
	assert(errno != EFAULT && "BUG: EFAULT from statfs/statvfs");
	switch (errno) {
	case EINTR: goto retry;
	case EIO: /* this is important enough to log: */
	case EOVERFLOW: /* this is important enough to log: */
		syslog(LOG_ERR, MOG_STR(MY_STATFS) "(%s) failed: %m", path);
	case ENOTDIR: /* race between read_file_system_list and statvfs */
	case ELOOP: /* race between read_file_system_list and statvfs? */
	case ENOSYS: /* if statvfs() doesn't work, fstatvfs() won't, either */
	case EACCES: /* this is common */
		return false;
	}
	/*
	 * assume other errors are recoverable (or fail elsewhere)
	 * ENAMETOOLONG - would fail anyways if we need to stat(2)
	 * ENOMEM - maybe the kernel will get more memory back, soon
	 */
	return true;
}
